#  Copyright (c) 2020. Davi Pereira dos Santos and Rafael Amatte Bisão
#  This file is part of the oka project.
#  Please respect the license. Removing authorship by any means
#  (by code make up or closing the sources) or ignoring property rights
#  is a crime and is unethical regarding the effort and time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
#
#  oka is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  oka is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with oka.  If not, see <http://www.gnu.org/licenses/>.
#
from dataclasses import dataclass
from typing import Union

import requests as req
from idict import idict
from idict.data.compression import unpack, pack
from idict.persistence.cache import Cache
from pandas import DataFrame

from oka.config import default_url


def j(r):
    """Helper function needed because flask test_client() provide json as a property(?), not as a method."""
    return r.json() if callable(r.json) else r.json


@dataclass
class Oka(Cache):
    """Client for OKA repository

    >>>
    Role:
        - Send/receive datasets or results as pandas DataFrame objects.
        - Send/receive datasets, processed data or results in any format (cannot be visualized on the web).
        - Seamless caching of data within a workflow.

    TODO:
        - Handle posts from command line to spare the user from accessing the web to publish results.
        - Distributed cache (after 'oka.lock(id, state)' is implemented to reserve a job and mark it is still alive).
    """

    # TODO: put sync here also
    token: str = None
    url: str = default_url
    id: str = None

    # Only used in parent class; useless here, but can be used for a Session
    decorator = None

    def lock(self, id, state):
        pass

    def __contains__(self, item):
        url = f"/api/item/{item}?checkonly=true"
        ret = j(self.request(url, "get"))
        return ret

    def __lshift__(self, other):
        """Can receive a Data object or a tuple (data, name, description)."""
        if isinstance(other, tuple):
            self.send(other[0], *other[1:])
        self.send(other)

    def __rmatmul__(self, other):
        return self.get(other)

    # noinspection PyArgumentList
    def __call__(self, id=None):
        if id is None:
            return self.__class__(url=self.url)
        return self.get(id)

    def request(self, route, method, **kwargs):
        headers = {"Authorization": "Bearer " + self.token}
        r = getattr(req, method)(self.url + route, headers=headers, **kwargs)
        if r.status_code == 401:
            raise Exception("Token invalid!")
        elif r.status_code == 422:
            pass
        else:
            if r.ok:
                return r
            print(r.content)
            print(j(r))
            print(j(r)["errors"])
            msg = j(r)["errors"]["json"]
            print(msg)
            raise Exception(msg)

    def get(self, id):
        """Get a dataframe, callable or idict from server

        On single-valued idicts, content (dataframe or callable) take precedence and is returned directly."""
        url = f"/api/item/{id}"
        if not (response := self.request(url, "get")):
            raise Exception(f"[Error] Data with OID {id} was not found.")
        value = unpack(response.content)
        if isinstance(value, DataFrame) or callable(value):
            return value
        dic = {}
        for k, v in value["ids"].items():
            url = f"/api/sync?id={id}&fetch=true"
            if not value:
                raise Exception(f"[Error] Missing key {k} for idict with OID {id}.")
            dic[k] = unpack(j(self.request(url, "get")))
        return idict(dic)
        # if not isinstance(value, dict) or "ids" not in value:
        #     raise Exception("dict containing key 'ids' expected.", type(value), value)

    def send(self, d: Union[DataFrame, idict], name=None, description=None):
        """Send a dataframe, callable or idict to server"""
        # Create inactive Post.
        # name = name or "→".join(
        #     x[:3] for x in data.history ^ "name" if x[:3] not in ["B", "Rev", "In", "Aut", "E"]
        # )
        if callable(d) or isinstance(d, DataFrame):
            d = idict(df=d)
        if name:
            d["_name"] = name
        if description:
            d["_description"] = description

        id = "_" + d.id[1:] if len(d.ids) == 1 else d.id
        url = f"/api/item/{id}"
        content = pack({"ids": d.ids})
        response = j(self.request(url, "post", files={"file": content}))["success"]
        if not response:
            print(f"Content already stored for id {d.id}")
            return d.id
        for k, v in d.ids.items():
            if k in d.blobs:
                # Use cached blob.
                content = d.blobs[k]
            else:
                content = pack(d[k], nondeterministic_fallback=True)
            url = f"/api/item/{v}"
            _ = j(self.request(url, "post", files={"file": content}))["success"]

        # except DuplicateEntryException as e:
        #     if "OKATESTING" not in os.environ:
        #         print(f"[Error] You have already uploaded the dataset with OID {e}.")

        return d.id

    def __setitem__(self, key, value):
        url = f"/api/item/{key}"
        content = pack(value)
        response = j(self.request(url, "post", files={"file": content}))["success"]
        if not response:
            print(f"Content already stored for id {key}")
            return None
        return response

    def __getitem__(self, key):
        url = f"/api/item/{key}"
        response = self.request(url, "get")
        if not response:
            raise Exception(f"[Error] Data with OID {key} was not found.")
        return unpack(response.content)

    def __delitem__(self, key):
        pass

    def __repr__(self):
        pass

    def __iter__(self):
        pass


# todo-icones/cores na web
# todo-checar mem leak
# todo-client enviar blob
# todo-enviar DF



import os

if "OKATESTING" in os.environ:
    # user:pass@http://localhost:5000
    default_url = "http://localhost:5000"  # os.environ['OKATESTING'].split("@")[1]
    default_user, default_password = os.environ["OKATESTING"].split("@")[0].split(":")
else:
    default_url = "https://oka.icmc.usp.br"
    default_user, default_password = None, None



from getpass import getpass

import numpy
import requests
from oka.config import default_url, default_user, default_password

from .client import Oka


def toy_df():
    from idict import idict

    return idict.fromtoy(output_format="df")["df"]


def j(r):
    """Helper function needed because flask test_client() provide json as a property(?), not as a method."""
    return r.json() if callable(r.json) else r.json


def generate_token(url=default_url):
    data = {
        "username": default_user or input("Username to connect to OKA: "),
        "password": default_password or getpass("Password to connect to OKA: "),
    }
    response = requests.post(f"{url}/api/auth/login", json=data)
    if response and "access_token" in j(response):
        return j(response)["access_token"]
    raise Exception("[Error] Authentication failed.")


def new_data(**kwargs):
    from aiuna import new

    for l in list(kwargs):
        if isinstance(kwargs[l], list):
            if not isinstance(kwargs[l][0], list):
                kwargs[l] = [[str(v)] for v in kwargs[l]]
            kwargs[l] = numpy.array(kwargs[l])
        metafield_d = f"{l.upper()}d"
        metafield_t = f"{l.upper()}t"
        if metafield_t not in kwargs:
            if len(kwargs[l].shape) > 1 and kwargs[l].shape[1] > 1:
                kwargs[metafield_d] = [f"{l}{i}" for i in range(len(kwargs[l][0]))]
                kwargs[metafield_t] = [f"real" for _ in kwargs[l][0]]
            else:
                kwargs[metafield_d] = [f"{l}{0}"]
                kwargs[metafield_t] = [list(set([str(v[0]) for v in kwargs[l]]))]
    return new(**kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['oka']

package_data = \
{'': ['*']}

install_requires = \
['idict>=1.211129.0,<2.0.0',
 'python-dotenv>=0.19.2,<0.20.0',
 'requests>=2.26.0,<3.0.0']

setup_kwargs = {
    'name': 'oka',
    'version': '0.211129.1',
    'description': 'Python client for oka repository',
    'long_description': '![test](https://github.com/rabizao/oka/workflows/test/badge.svg)\n[![codecov](https://codecov.io/gh/rabizao/oka/branch/main/graph/badge.svg)](https://codecov.io/gh/rabizao/oka)\n<a href="https://pypi.org/project/oka">\n<img src="https://img.shields.io/pypi/v/oka.svg?label=release&color=blue&style=flat-square" alt="pypi">\n</a>\n![Python version](https://img.shields.io/badge/python-3.8%20%7C%203.9-blue.svg)\n[![license: GPL v3](https://img.shields.io/badge/License-GPLv3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)\n\n[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.5501845.svg)](https://doi.org/10.5281/zenodo.5501845)\n[![arXiv](https://img.shields.io/badge/arXiv-2109.06028-b31b1b.svg?style=flat-square)](https://arxiv.org/abs/2109.06028)\n[![User Manual](https://img.shields.io/badge/doc-user%20manual-a0a0a0.svg)](https://rabizao.github.io/oka)\n[![API Documentation](https://img.shields.io/badge/doc-API%20%28auto%29-a0a0a0.svg)](https://rabizao.github.io/oka/api)\n\n# oka - Client for OKA repository\n[Latest version as a package](https://pypi.org/project/oka)\n\n[Current code](https://github.com/rabizao/oka)\n\n[User manual](https://rabizao.github.io/oka)\n\n[API documentation](https://rabizao.github.io/oka/api)\n\n## Overview\n`oka` is a client for Oka repository.\nIt also provides utilities to process data.\n\n## Installation\n### ...as a standalone lib\n```bash\n# Set up a virtualenv. \npython3 -m venv venv\nsource venv/bin/activate\n\n# Install from PyPI...\npip install --upgrade pip\npip install -U oka\npip install -U oka[full]  # use the flag \'full\' for extra functionality (recommended)\n\n# ...or, install from updated source code.\npip install git+https://github.com/rabizao/oka\n```\n    \n### ...from source\n```bash\nsudo apt install python3.8-venv python3.8-dev python3.8-distutils # For Debian-like systems.\ngit clone https://github.com/rabizao/oka\ncd oka\npython3.8 -m venv venv\nsource venv/bin/activate\npip install -e .\n```\n\n## Usage\n\n\n\n\n\n\n**Hello world**\n<details>\n<p>\n\n```python3\nfrom oka import Oka, generate_token, toy_df\n\n# Create a pandas dataframe.\ndf = toy_df()\nprint(df.head())\n"""\n   attr1  attr2  class\n0    5.1    6.4      0\n1    1.1    2.5      1\n2    6.1    3.6      0\n3    1.1    3.5      1\n4    3.1    2.5      0\n"""\n```\n\n```python3\n\n# Login.\ntoken = generate_token("http://localhost:5000")\nclient = Oka(token, "http://localhost:5000")\n\n# Store.\nid = client.send(df)\n\n# Store again.\nid = client.send(df)\n"""\nContent already stored for id iJ_e4463c51904e9efb800533d25082af2a7bf77\n"""\n\n# Fetch.\ndf = client.get(id)\n\nprint(df.head())\n"""\n   attr1  attr2  class\n0    5.1    6.4      0\n1    1.1    2.5      1\n2    6.1    3.6      0\n3    1.1    3.5      1\n4    3.1    2.5      0\n"""\n```\n\n</p>\n</details>\n\n\n\n\n\n\n\n\n\n**DataFrame by hand**\n<details>\n<p>\n\n```python3\nimport pandas as pd\nfrom oka import Oka, generate_token\n\n# Create a pandas dataframe.\ndf = pd.DataFrame(\n    [[1, 2, "+"],\n     [3, 4, "-"]],\n    index=["row 1", "row 2"],\n    columns=["col 1", "col 2", "class"],\n)\nprint(df.head())\n"""\n       col 1  col 2 class\nrow 1      1      2     +\nrow 2      3      4     -\n"""\n```\n\n```python3\n\n# Login.\ntoken = generate_token("http://localhost:5000")\nclient = Oka(token, "http://localhost:5000")\n\n# Store.\nid = client.send(df)\n\n# Store again.\nid = client.send(df)\n"""\nContent already stored for id f7_6b9deafec2562edde56bfdc573b336b55cb16\n"""\n\n# Fetch.\ndf = client.get(id)\n\nprint(df.head())\n"""\n       col 1  col 2 class\nrow 1      1      2     +\nrow 2      3      4     -\n"""\n```\n\n\n\n\n\n\n**Machine Learning workflow**\n<details>\n<p>\n\n```python3\nfrom pprint import pprint\n\nfrom idict import let, idict\nfrom idict.function.classification import fit, predict\nfrom idict.function.evaluation import split\nfrom sklearn.ensemble import RandomForestClassifier as RF\n\nd = (\n        idict.fromtoy()\n        >> split\n        >> let(fit, algorithm=RF, config={"n_estimators": 55}, Xin="Xtr", yin="ytr")\n        >> let(predict, Xin="Xts")\n)\n\nprint(d.z)\n"""\n[1 0 1 0 1 1 1]\n"""\n```\n\n```python3\n\npprint(d.history)\n"""\n{\'fit--------------------------------idict\': {\'code\': \'def f(algorithm=None, \'\n                                                      "config={}, Xin=\'X\', "\n                                                      "yin=\'y\', "\n                                                      "output=\'model\', "\n                                                      \'version=0, **kwargs):\\n\'\n                                                      \'obj = \'\n                                                      \'algorithm(**config)\\n\'\n                                                      \'obj.fit(kwargs[Xin], \'\n                                                      \'kwargs[yin])\\n\'\n                                                      \'return {output: obj, \'\n                                                      "\'_history\': ...}",\n                                              \'description\': \'Induce a model.\',\n                                              \'name\': \'fit\',\n                                              \'parameters\': {\'Xin\': \'Xtr\',\n                                                             \'algorithm\': <class \'sklearn.ensemble._forest.RandomForestClassifier\'>,\n                                                             \'config\': {\'n_estimators\': 55},\n                                                             \'output\': \'model\',\n                                                             \'version\': 0,\n                                                             \'yin\': \'ytr\'}},\n \'predict----------------------------idict\': {\'code\': "def f(input=\'model\', "\n                                                      "Xin=\'X\', yout=\'z\', "\n                                                      \'version=0, **kwargs):\\n\'\n                                                      \'return {yout: \'\n                                                      \'kwargs[input].predict(kwargs[Xin]), \'\n                                                      "\'_history\': ...}",\n                                              \'description\': \'Predict values \'\n                                                             \'according to a \'\n                                                             \'model.\',\n                                              \'name\': \'predict\',\n                                              \'parameters\': {\'Xin\': \'Xts\',\n                                                             \'input\': \'model\',\n                                                             \'version\': 0,\n                                                             \'yout\': \'z\'}},\n \'split----------------------sklearn-1.0.1\': {\'code\': "def f(input=[\'X\', \'y\'], "\n                                                      "config={\'test_size\': "\n                                                      "0.33, \'shuffle\': True, "\n                                                      "\'stratify\': \'y\', "\n                                                      "\'random_state\': 0}, "\n                                                      \'**kwargs):\\n\'\n                                                      "if input != [\'X\', "\n                                                      "\'y\']:\\n"\n                                                      \'    raise \'\n                                                      \'Exception(f"Not \'\n                                                      \'implemented for \'\n                                                      \'input/output different \'\n                                                      \'from default values: \'\n                                                      \'{input}")\\n\'\n                                                      \'from \'\n                                                      \'sklearn.model_selection \'\n                                                      \'import \'\n                                                      \'train_test_split\\n\'\n                                                      \'args = {}\\n\'\n                                                      \'for i, _ in \'\n                                                      \'enumerate(input):\\n\'\n                                                      \'    args[input[i]] = \'\n                                                      \'kwargs[input[i]]\\n\'\n                                                      \'else:\\n\'\n                                                      "    if \'stratify\' in "\n                                                      \'config:\\n\'\n                                                      \'        if \'\n                                                      "isinstance(config[\'stratify\'], "\n                                                      \'str):\\n\'\n                                                      \'            if \'\n                                                      "config[\'stratify\'] not "\n                                                      \'in input:\\n\'\n                                                      \'                raise \'\n                                                      \'Exception(f"Missing \'\n                                                      \'field \'\n                                                      "{config[\'stratify\']} "\n                                                      \'for stratification.")\\n\'\n                                                      \'            \'\n                                                      "config[\'stratify\'] = "\n                                                      "args[config[\'stratify\']]\\n"\n                                                      \'    Xtr, Xts, ytr, yts \'\n                                                      \'= \'\n                                                      \'train_test_split(*(args.values)(), \'\n                                                      \'**config)\\n\'\n                                                      "    return {\'Xtr\':Xtr,  "\n                                                      "\'ytr\':ytr,  \'Xts\':Xts,  "\n                                                      "\'yts\':yts,  "\n                                                      "\'_history\':...}",\n                                              \'description\': \'Split data in \'\n                                                             \'two sets.\',\n                                              \'name\': \'split\',\n                                              \'parameters\': {\'config\': {\'random_state\': 0,\n                                                                        \'shuffle\': True,\n                                                                        \'stratify\': array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1]),\n                                                                        \'test_size\': 0.33},\n                                                             \'input\': [\'X\',\n                                                                       \'y\']}}}\n"""\n```\n\n\n</p>\n</details>\n\n\n\n\n\n\n\n## More info\nAside from the papers on [identification](https://arxiv.org/abs/2109.06028)\nand on [similarity (not ready yet)](https://), the [PyPI package](https://pypi.org/project/oka) \nand [GitHub repository](https://github.com/davips/rabizao/oka), \n<!-- one can find more information, at a higher level application perspective,  -->\nA lower level perspective is provided in the [API documentation](https://rabizao.github.io/oka).\n\n\n\n## Grants\nThis work was supported by Fapesp under supervision of\nProf. André C. P. L. F. de Carvalho at CEPID-CeMEAI (Grants 2013/07375-0 – 2019/01735-0).\n\n.>>>>>>>>> outros <<<<<<<<<<<.\n',
    'author': 'Rafael Bizao',
    'author_email': 'rabizao@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
